/*****************************************************************************/
/* Copyright (C) 2003-2009, JOANNEUM RESEARCH Forschungs Ges.m.b.H.          */
/* Steyrergasse 17, A-8010 Graz, Austria.                                    */
/*                                                                           */
/*  This file is part of the JRS MPEG-7 Library.                             */
/*  http://mpeg-7.joanneum.at                                                */
/*                                                                           */
/*  JRS MPEG-7 Library is free software: you can redistribute it and/or      */
/*  modify it under the terms of the GNU Lesser General Public License       */
/*  (LGPL) as published by the Free Software Foundation, either version 3    */
/*  of the License, or (at your option) any later version.                   */
/*                                                                           */
/*  JRS MPEG-7 Library is distributed in the hope that it will be useful,    */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of           */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the             */
/*  GNU Lesser General Public License for more details.                      */
/*                                                                           */
/*  You should have received a copy of the GNU Lesser General Public License */
/*  along with this library. If not, see <http://www.gnu.org/licenses/>.     */
/*****************************************************************************/



#ifndef MP7JRSARCHIVE_H
#define MP7JRSARCHIVE_H

#pragma once

#include "Mp7JrsNode.h"
#include "Mp7JrsPtr.h"
#include "Mp7JrsDefines.h"
#include "Mp7JrsNodeCollection.h"
#include <xercesc/util/XMLString.hpp>
#include <xercesc/framework/XMLFormatter.hpp>

#include "Mp7JrsSchemaLocationMap.h"
#include "Mp7JrsException.h"

/** Mp7Jrs archive.
  * A class which handles (de)serializing of Mp7Jrs XML documents.
  */
class MP7JRS_EXPORT Mp7JrsArchive
{

public:
	/** @name Constructors and Destructors
	 */
	//@{

	Mp7JrsArchive(void);
	~Mp7JrsArchive(void);

	//@}

	/** @name Serializing and Deserializing
	 */
	//@{

	/** Serializes to a file.
	  * @param path Path and file name of the destination file.
	  * @param node Root node of the subtree to be serialized.
	  * @param nodeName Name of the root node (only necessary if a node of this type may have different names when used in different places of the schema).
	  * @param useTypeAttr Write the xsi:type attribute for all nodes in the subtree.
	  */
	void ToFile(const XMLCh * path, Mp7JrsNodePtr &node, XMLCh* nodeName = (XMLCh*)NULL, bool useTypeAttr = false);

	/** Serializes to a file.
	  * @param path Path and file name of the destination file.
	  * @param node Root node of the subtree to be serialized.
	  * @param nodeName Name of the root node (only necessary if a node of this type may have different names when used in different places of the schema).
	  * @param useTypeAttr Write the xsi:type attribute for all nodes in the subtree.
	  */
	void ToFile(const char* path, Mp7JrsNodePtr &node, XMLCh* nodeName = (XMLCh*)NULL, bool useTypeAttr = false);
	
	// TODO FTT Maybe we shouldn't use optional args for similar code in C# and JAVA

	/** Deserializes from file.
	  * @param path Path and file name of the source file.
	  * @param node Node to be used as parent of the fragement to be deserialized.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromFile(const XMLCh * path, Mp7JrsNodePtr &node);

	/** Deserializes from file.
	  * The root node of the fragment in the file will become the root node of the document.
	  * @param path Path and file name of the source file.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromFile(const XMLCh * path);

	/** Deserializes from file.
	  * @param path Path and file name of the source file.
	  * @param node Node to be used as parent of the fragement to be deserialized.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromFile(const char * path, Mp7JrsNodePtr &node);

	/** Deserializes from file.
	  * The root node of the fragment in the file will become the root node of the document.
	  * @param path Path and file name of the source file.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromFile(const char * path);

//---

	/** Constants that describe how to add something to a collection.
	 * - APPEND: append to the end of the collection.
	 * - INSERT: insert at given index.
	 * - REPLACE: replace element at given index.
	 */
	enum HowToAdd { APPEND, INSERT, REPLACE };

	/** Inserts a fragment from file into a collection.
	  * @param path Path and file name of the source file.
	  * @param collection Collection where the fragment is to be added.
	  * @param howtoadd how to add the fragment.
	  * @param where the index where the fragment is to be added.
	  * @return The index of the added node, or -1 on failure.
	  */
	int AddToCollectionFromFile(const XMLCh * path, Mp7JrsNodeCollectionPtr &collection,
		HowToAdd howtoadd, int where = -1);

	/** Inserts a fragment from a buffer into a collection.
	  * @param buffer Buffer containing the MPEG-7 XML.
	  * @param collection Collection where the fragment is to be added.
	  * @param howtoadd how to add the fragment.
	  * @param where the index where the fragment is to be added.
	  * @return The index of the added node, or -1 on failure.
	  */
	int AddToCollectionFromBuffer(const XMLCh * buffer, Mp7JrsNodeCollectionPtr &collection,
		HowToAdd howtoadd, int where = -1);

	/** Inserts a fragment from file into a collection.
	  * @param path Path and file name of the source file.
	  * @param collection Collection where the fragment is to be added.
	  * @param howtoadd how to add the fragment.
	  * @param where the index where the fragment is to be added.
	  * @return The index of the added node, or -1 on failure.
	  */
	int AddToCollectionFromFile(const char * path, Mp7JrsNodeCollectionPtr &collection,
		HowToAdd howtoadd, int where = -1);
		
	/** Inserts a fragment from a buffer into a collection.
	  * @param buffer Buffer containing the MPEG-7 XML.
	  * @param collection Collection where the fragment is to be added.
	  * @param howtoadd how to add the fragment.
	  * @param where the index where the fragment is to be added.
	  * @return The index of the added node, or -1 on failure.
	  */
	int AddToCollectionFromBuffer(const char * buffer, Mp7JrsNodeCollectionPtr &collection,
		HowToAdd howtoadd, int where = -1);
//---

	/** Serializes to a 16 bit Unicode (Xerces) buffer.
	  * @param node Root node of the subtree to be serialized.
	  * @param nodeName Name of the root node (only necessary if a node of this type may have different names when used in different places of the schema).
	  * @param useTypeAttr Write the xsi:type attribute for all nodes in the subtree.
	  * @return a 16 bit Unicode buffer (on the Xerces heap), which *MUST* be released via <code>DeleteBuffer()</code>
	  */
	XMLCh* ToWBuffer(Mp7JrsNodePtr &node, XMLCh* nodeName = (XMLCh*)NULL, bool useTypeAttr = false);

	/** Serializes to a 8 bit (Xerces) buffer.
	  * @param node Root node of the subtree to be serialized.
	  * @param nodeName Name of the root node (only necessary if a node of this type may have different names when used in different places of the schema).
	  * @param useTypeAttr Write the xsi:type attribute for all nodes in the subtree.
	  * @return a 8 bit buffer (on the Xerces heap), which *MUST* be released via <code>DeleteBuffer()</code>
	  */
	char* ToBuffer(Mp7JrsNodePtr &node, XMLCh* nodeName = (XMLCh*)NULL, bool useTypeAttr = false);

	/** Deserializes from a 16 bit Unicode buffer.
	  * @param buf Source buffer.
	  * @param node Node to be used as parent of the fragement to be deserialized.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromBuffer(const XMLCh * buf, Mp7JrsNodePtr &node);

	/** Deserializes from a 16 bit Unicode buffer.
	  * The root node of the fragment in the buffer will become the root node of the document.
	  * @param buf Source buffer.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromBuffer(const XMLCh * buf);

	/** Deserializes from a 8 bit buffer.
	  * @param buf Source buffer.
	  * @param node Node to be used as parent of the fragement to be deserialized.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromBuffer(const char * buf, Mp7JrsNodePtr &node);

	/** Deserializes from a 8 bit buffer.
	  * The root node of the fragment in the buffer will become the root node of the document.
	  * @param buf Source buffer.
	  * @return The root node of the deserialized fragment.
	  */
	Mp7JrsNodePtr FromBuffer(const char * buf);

	/** Serializes to a console (stdout).
	  * @param node Root node of the subtree to be serialized.
	  * @param nodeName Name of the root node (only necessary if a node of this type may have different names when used in different places of the schema).
	  * @param useTypeAttr Write the xsi:type attribute for all nodes in the subtree.
	  */
	void ToStdOut(Mp7JrsNodePtr &node, XMLCh* nodeName = (XMLCh*)NULL, bool useTypeAttr = false);

	/** Deletes a buffer created when serializing (on the Xerces heap). Note: if you call this method with a pointer to your heap, the system becomes inconsitent.
	  * @param buffer Buffer to be deleted. 
	  */
	void DeleteBuffer(char** buffer);

	/** Deletes a buffer created when serializing (on the Xerces heap). Note: if you call this method with a pointer to your heap, the system becomes inconsitent.
	  * @param buffer Buffer to be deleted. 
	  */
	void DeleteBuffer(XMLCh** buffer);

	//@}

	/** @name Get and Set Options
	 */
	//@{

	/** Set the encoding (UTF-8, UTF-16 ...) of XML document to be (de)serialized. */
	void SetEncoding(const char * encoding) { SetEncoding(X(encoding)); } 
	/** Set the encoding (UTF-8, UTF-16 ...) of XML document to be (de)serialized. */
	void SetEncoding(const XMLCh* encoding);

	/** Get current encoding setting (UTF-8, UTF-16 ...) */
	char * GetEncoding();

	/** Set a schema location.
	  * @param namespaceURI URI of the schema namespace.
	  * @param prefix Namespace prefix of the schema.
	  * @param schemaLocation Local path and file name of the schema.
	  */
	void SetSchemaLocation(const char* namespaceURI, const char* prefix, const char * schemaLocation) { SetSchemaLocation(X(namespaceURI),X(prefix),X(schemaLocation)); } 

	/** Set a schema location.
	  * @param namespaceURI URI of the schema namespace.
	  * @param prefix Namespace prefix of the schema.
	  * @param schemaLocation Local path and file name of the schema.
	  */
	void SetSchemaLocation(const XMLCh* namespaceURI, const XMLCh* prefix, const XMLCh * schemaLocation);

	/** Get location of given schema.
	  */
	char * GetSchemaLocation(const char* namespaceURI);

	/** Sets the default namespace. If there is an existing entry it will be overwritten. 
	  * @param namespaceURI Namespace name
	  * @param prefix default namespace prefix, if used, otherwise maybe 0
	  */
	void SetDefaultNamespace(const XMLCh* namespaceURI, const XMLCh* prefix);

	/** Gets the namespace's name and prefix of the default namespace.
	  * @param namespaceURI Default namespace name
	  * @param prefix Default namespace prefix, if used, otherwise maybe null
	  * @return true if values set in the arguments, false if no namespace has been found. The prefix can be null.
	  */
	bool GetDefaultNamespace(const XMLCh* * namespaceURI, const XMLCh* * prefix);


	/** Set if validation shall be done
	  * @param validate If true, validation will be done.
	  */
	void SetDoValidation(bool validate) { this->validate = validate; }
	
	/** Get validation setting. */
	bool GetDoValidation() { return validate; }

	/** Set if full schema checking shall be done.
	  * @param fullSchema If true, full schema checking will be done.
	  * <b>IMPORTANT NOTE: This will cause Xerces 2.2 to crash at large maxOccurs values. Do not use!</b>
	  */
	void SetFullSchemaChecking(bool fullSchema) { this->fullSchema = fullSchema; }
	
	/** Get full schema checking setting */
	bool GetFullSchemaChecking() { return fullSchema; }
	
	/** Set if pretty formatted (with line breaks) output is desired.
	  */
	void SetPrettyFormatting(bool flag) { this->prettyformat = flag; }
	
	/** Get if pretty formatted output flag (using line breaks) is set.
	bool GetPrettyFormatting() { return prettyformat; }
	
	
	/** Set whether the XMLDecl header (version, encoding, standalone) shall be written.
	  * @param decl If true, the XMLDecl header will be written when serializing. */
	void SetWriteXMLDecl(bool decl) { writeXMLDecl = decl; xmlDeclDefault = false; }

	/** Get XMLDecl setting */
	bool GetWriteXMLDecl() { return writeXMLDecl; }

	/** Set whether the namespace declarations should be written to the XML root node or not (default: true).
	  * This is necessary to avoid xsi namespace declarations in case of an xsi:type all over the place.
	  * @param flag If true, all namespaces are declared in the XML root node. */
	void SetForceNamespaceDeclaration(bool flag) { forceNamespaceDeclaration = flag; }

	/** Get ForceNamespaceDeclaration setting */
	bool GetForceNamespaceDeclaration() { return forceNamespaceDeclaration; }

	/** Set whether or not you want your elements of the default namespace be prefixed (Initially: false).
	  * @param flag If true, xmlns:mpeg7=urn:mpeg:mpeg7:schema:2004 is used instead of xmlns=urn:mpeg:mpeg7:schema:2004
	  */
	void SetUsePrefixForDefaultNS(bool flag) { this->usePrefixForDefaultNS = flag; }

	/** Get whether or not you want your elements of the default namespace be prefixed with mpeg7 (Initially: false). 
	*/
	bool GetUsePrefixForDefaultNS() { return this->usePrefixForDefaultNS; }
	
	//@}

private:

	class XercesWrapper;
	friend class XercesWrapper;
	
	void _Serialize(Mp7JrsNodePtr &node, XMLCh* nodeName = (XMLCh*)NULL, bool useTypeAttr = false, bool writeHeader = true);

	int _DeserializeIntoCollection(InputSource &source,
	Mp7JrsNodeCollectionPtr &collection, HowToAdd howtoadd, int where);

	Mp7JrsNodePtr _Deserialize(Mp7JrsNodePtr &node,
		InputSource &source);

	
	XMLCh* encoding;
	XMLFormatTarget* destination;
	bool standalone;
	XMLCh* version;

	Mp7JrsSchemaLocationMap	_locationMap;

	bool validate;			// do validation
	bool fullSchema;		// do full schema checking
	bool writeXMLDecl;		// write XMLDecl
	bool xmlDeclDefault;	// use default setting for writing XMLDecl
							// default=true for complete files and false for fragments
							// default is used, if SetXMLDecl has not been called
							
	bool usePrefixForDefaultNS; // If true, all elements of the default namespace are prefixed. If no prefix is available, the prefix is ignored
	bool prettyformat; // Readable vs small xmlfiles
	bool forceNamespaceDeclaration;		// Write namespace decls in XML root node

	static Mp7JrsClientID maxId; //maximum client ID
};

#endif
